"""Module for printing a block of text overwriting previous text."""

import sys
from threading import RLock


class Printer:
    """Reprints messages wiping unneeded lines. Supports multiple threads."""

    __rlock = RLock()

    def __init__(self, line_length=70):
        """Initialize object."""
        self.__msgs = {}
        self.__line_length = line_length

    def add_msg(self, key, msg):
        """Add a new message and print it on last line."""
        with self.__rlock:
            self.__msgs[key] = msg
            print(self.__msgs[key].ljust(self.__line_length, " "))

    def print_msg(self, msg):
        """Print a single message."""
        print(msg.ljust(self.__line_length, " "))

    def purge_msg(self, key, last_msg):
        """Print the last message on top active line and move lower."""
        with self.__rlock:
            self.__move_up()
            if key in self.__msgs:
                del self.__msgs[key]
            print(last_msg.ljust(self.__line_length, " "))
            self.__print_active(move_up=False)

    def __print_active(self, move_up=False):
        """Print all active messages overwriting console."""
        # Clear previous text by overwritig non-spaces with spaces
        if move_up:
            self.__move_up()
        for key in self.__msgs.keys():
            print(self.__msgs[key].ljust(self.__line_length, " "))

    def __move_up(self):
        """Move cursor to the top active line."""
        for _ in range(len(self.__msgs)):
            sys.stdout.write("\033[A")



"""Summary.

Attributes:
    log (TYPE): Description
"""
import logging
import subprocess
from os import path
from termcolor import colored
from concurrent import futures

from catkin_tools_fetch.lib.tools import Tools
from catkin_tools_fetch.lib.tools import GitBridge
from catkin_tools_fetch.lib.printer import Printer

log = logging.getLogger('deps')


class Updater(object):
    """Updater class. Handles the updating of all packages."""

    PULLED_TAG = "[PULLED]"
    UP_TO_DATE_TAG = "[UP TO DATE]"
    OK_TAGS = [PULLED_TAG, UP_TO_DATE_TAG]

    CHANGES_TAG = "[UNCOMMITTED CHANGES]"
    RUNNING_TAG = "[RUNNING]"
    NO_TRACK_TAG = "[NO BRANCH]"
    ERROR_TAG = "[GIT ERROR]"
    CONFLICT_TAG = "[MERGE CONFLICT]"

    UP_TO_DATE_MSG = "Already up-to-date"
    CONFLICT_MSG = "Automatic merge failed"

    def __init__(self,
                 ws_path,
                 packages,
                 conflict_strategy,
                 use_preprint=True,
                 colored=True,
                 num_threads=4):
        """Initialize the updater.

        Args:
            ws_path (str): Path to the workspace
            packages (dict(str)): Dictionary of packages to be downloaded
            conflict_strategy (str): A strategy to handle conflicts
        """
        super(Updater, self).__init__()
        self.ws_path = ws_path
        self.packages = packages
        self.conflict_strategy = conflict_strategy
        self.thread_pool = futures.ThreadPoolExecutor(max_workers=num_threads)
        self.printer = Printer()
        self.colored = colored
        self.use_preprint = use_preprint

    def filter_packages(self, selected_packages):
        """Filter the packages based on user input.

        Args:
            selected_packages (str[]): List of package names picked by the user

        Returns:
            dict: A dictionary of packages that match selected_packages.
        """
        if not selected_packages:
            return self.packages
        filtered_packages = {}
        for ws_folder, package in self.packages.items():
            if package.name in selected_packages:
                filtered_packages[ws_folder] = package
        return filtered_packages

    def pick_tag(self, folder, package):
        """Pick result tag for a folder."""
        if self.use_preprint:
            msg = " {}: {}".format(Tools.decorate(
                package.name), Updater.RUNNING_TAG)
            self.printer.add_msg(package.name, msg)
        output, branch, has_changes = GitBridge.status(folder)
        if has_changes:
            return package, Updater.CHANGES_TAG
        try:
            output = GitBridge.pull(folder, branch)
            return package, Updater.tag_from_output(output)
        except subprocess.CalledProcessError as e:
            log.debug(" git pull returned error: %s", e)
            return package, Updater.ERROR_TAG

    def update_packages(self, selected_packages):
        """Update all the folders to match the remote. Considers the branch.

        Args:
            selected_packages (str[]): List of packages picked by the user.

        Returns:
            status_msgs (list(tuple)): return a list of tupples (pkg_name, tag)
        """
        log.info(" Pulling packages:")
        packages = self.filter_packages(selected_packages)
        status_msgs = []
        futures_list = []
        for ws_folder, package in packages.items():
            picked_tag = None
            folder = path.join(self.ws_path, ws_folder)
            futures_list.append(
                self.thread_pool.submit(self.pick_tag, folder, package))
        for future in futures.as_completed(futures_list):
            package, picked_tag = future.result()
            # change logger for warning if something is wrong
            if self.colored:
                picked_tag = Updater.colorize_tag(picked_tag)
            # now show the results to the user
            status_msgs.append((package.name, picked_tag))
            msg = " {}: {}".format(Tools.decorate(package.name), picked_tag)
            self.printer.purge_msg(package.name, msg)
        return status_msgs

    @staticmethod
    def tag_from_output(output):
        """Get tag from output."""
        try:
            str_output = output.decode("utf-8")
        except AttributeError:
            str_output = output
        if Updater.UP_TO_DATE_MSG in str_output:
            return Updater.UP_TO_DATE_TAG
        if Updater.CONFLICT_MSG in str_output:
            return Updater.CONFLICT_TAG
        return Updater.PULLED_TAG

    @staticmethod
    def colorize_tag(picked_tag):
        """Colorize the tag."""
        if picked_tag in Updater.OK_TAGS:
            return colored(picked_tag, 'green')

        if picked_tag == Updater.CHANGES_TAG:
            # this is a warning
            return colored(picked_tag, 'yellow')
        return colored(picked_tag, 'red')


class Strategy(object):
    """An enum of stategies for update."""

    IGNORE = 'ignore'
    ABORT = 'abort'
    # STASH = 'stash'

    @classmethod
    def list_all(cls):
        """List all strategies."""
        all_strategies = [val for key, val in cls.__dict__.items()
                          if not callable(getattr(cls, key))
                          and not key.startswith("__")]
        return all_strategies



"""Contains main for deps verb.

Attributes:
    log (logging.Log): logger
"""

import sys
import logging
from os import path
from argparse import ArgumentParser

try:
    from catkin_pkg.packages import find_packages
except ImportError as e:
    sys.exit(
        'ImportError: "from catkin_pkg.topological_order import '
        'topological_order" failed: %s\nMake sure that you have installed '
        '"catkin_pkg", and that it is up to date and on the PYTHONPATH.' % e
    )

from catkin_tools.argument_parsing import add_context_args
from catkin_tools.context import Context

from catkin_tools_fetch.lib.dependency_parser import Parser
from catkin_tools_fetch.lib.downloader import Downloader
from catkin_tools_fetch.lib.tools import Tools
from catkin_tools_fetch.lib.update import Updater
from catkin_tools_fetch.lib.update import Strategy

logging.basicConfig()
log = logging.getLogger('deps')


def prepare_arguments(parser):
    """Parse arguments that belong to this verb.

    Args:
        parser (argparser): Argument parser

    Returns:
        argparser: Parser that knows about our flags.
    """
    parser.description = """ Download dependencies for one or more packages in
        a catkin workspace. This reads dependencies from package.xml file of
        each of the packages in the workspace and tries to download their
        sources from version control system of choice."""
    add_context_args(parser)

    packages_help_msg = """
        Packages for which the dependencies are analyzed.
        If no packages are given, all packages are processed."""
    fetch_group = parser.add_argument_group(
        'Packages',
        'Control for which packages we fetch dependencies.')

    fetch_group.add_argument('packages',
                             metavar='PKGNAME',
                             nargs='*',
                             help=packages_help_msg)

    # add config flags to all groups that need it
    config_group = parser.add_argument_group('Config')
    config_group.add_argument(
        '--default_url', default="{package}",
        help='Where to look for packages by default.')
    config_group.add_argument('--no_status', action='store_true',
                              help='Do not use progress status when cloning.')
    config_group.add_argument('--num_threads', '-j',
                              type=int,
                              default=4,
                              help='Number of threads run in parallel.')

    # Behavior
    behavior_group = parser.add_argument_group(
        'Interface', 'The behavior of the command-line interface.')
    add = behavior_group.add_argument
    add('--verbose', '-v', action='store_true', default=False,
        help='Print output from commands.')

    return parser


def prepare_arguments_deps(parser):
    """Parse arguments that belong to this verb.

    Args:
        parser (argparser): Argument parser

    Returns:
        argparser: Parser that knows about our flags.
    """
    parser.description = """ Manage dependencies for one or more packages in
        a catkin workspace. This reads dependencies from package.xml file of
        each of the packages in the workspace and tries to download their
        sources from version control system of choice."""
    add_context_args(parser)

    parent_parser = ArgumentParser(add_help=False)

    # add config flags to all groups that need it
    parent_parser.add_argument(
        '--default_url', default="{package}",
        help='Where to look for packages by default.')

    # Behavior
    parent_parser.add_argument('--verbose', '-v',
                               action='store_true',
                               default=False,
                               help='Print output from commands.')
    parent_parser.add_argument('--no_status',
                               action='store_true',
                               default=False,
                               help='Do not use progress status when cloning.')
    parent_parser.add_argument('--num_threads', '-j',
                               type=int,
                               default=4,
                               help='Number of threads run in parallel.')

    packages_help_msg = """
        Packages for which the dependencies are analyzed.
        If no packages are given, all packages are processed."""

    # we need subparsers for this verb
    subparsers = parser.add_subparsers(dest='subverb', help="Possible verbs.")

    # add a parser for update sub-verb
    update_help_msg = """
        Update the existing repositories to their latest state from remote."""
    parser_update = subparsers.add_parser(
        'update', help=update_help_msg, parents=[parent_parser])
    config_update_group = parser_update.add_argument_group('Config')
    conflict_help_msg = """ When we pull a git repository there can be
        conflicts. We need to resolve them in some way. You can pick this here.
        By default the plugin will use '%(default)s' strategy."""
    config_update_group.add_argument('--on-conflict', '-r',
                                     choices=Strategy.list_all(),
                                     default=Strategy.IGNORE,
                                     help=conflict_help_msg)

    update_pkg_group = parser_update.add_argument_group(
        'Packages',
        'Control for which packages we update dependencies.')
    update_pkg_group.add_argument('packages',
                                  metavar='PKGNAME',
                                  nargs='*',
                                  help=packages_help_msg)

    # add a parser for fetch sub-verb
    fetch_help_msg = """
        Fetch the dependencies stored package.xml files."""
    parser_fetch = subparsers.add_parser('fetch',
                                         help=fetch_help_msg,
                                         parents=[parent_parser])
    fetch_group = parser_fetch.add_argument_group(
        'Packages',
        'Control for which packages we fetch dependencies.')
    fetch_group.add_argument('packages',
                             metavar='PKGNAME',
                             nargs='*',
                             help=packages_help_msg)

    return parser


def main(opts):
    """Main function for fetch verb.

    Args:
        opts (dict): Options populated by an arg parser.

    Returns:
        int: Return code
    """
    # Load the context
    if opts.verbose:
        log.setLevel(logging.getLevelName("DEBUG"))
        log.debug(" Enabling DEBUG output.")
    else:
        log.setLevel(logging.getLevelName("INFO"))

    if opts.no_status:
        log.info(" Not printing status messages while cloning.")
        use_preprint = False
    else:
        log.info(" Will print status messages while cloning.")
        use_preprint = True

    log.info(" Using %s threads.", opts.num_threads)

    context = Context.load(opts.workspace, opts.profile, opts, append=True)
    default_url = Tools.prepare_default_url(opts.default_url)
    if not opts.workspace:
        log.critical(" Workspace undefined! Abort!")
        return 1
    if opts.verb == 'fetch' or opts.subverb == 'fetch':
        if opts.verb == 'fetch':
            msg = """
############################# DEPRECATED ###################################
You are using an old deprecated command: `catkin fetch`.
Please use the new command instead:

                `catkin fetch` --> `catkin deps fetch`

It has the same interface. The old command will be removed in future.
############################################################################
            """
            log.warning(msg)
            seconds_to_sleep = 5
            log.info(" Showing deprecation banner for %s seconds.",
                     seconds_to_sleep)
            from time import sleep
            sleep(seconds_to_sleep)
        return fetch(packages=opts.packages,
                     workspace=opts.workspace,
                     context=context,
                     default_url=default_url,
                     use_preprint=use_preprint,
                     num_threads=opts.num_threads)
    if opts.subverb == 'update':
        return update(packages=opts.packages,
                      workspace=opts.workspace,
                      context=context,
                      default_url=default_url,
                      conflict_strategy=opts.on_conflict,
                      use_preprint=use_preprint,
                      num_threads=opts.num_threads)


def update(packages,
           workspace,
           context,
           default_url,
           conflict_strategy,
           use_preprint,
           num_threads):
    """Update packages from the available remotes.

    Args:
        packages (list): A list of packages provided by the user.
        workspace (str): Path to a workspace (without src/ in the end).
        context (Context): Current context. Needed to find current packages.
        default_url (str): A default url with a {package} placeholder in it.
        use_preprint (bool): Show status messages while cloning

    Returns:
        int: Return code. 0 if success. Git error code otherwise.
    """
    ws_path = path.join(workspace, 'src')
    workspace_packages = find_packages(context.source_space_abs,
                                       exclude_subspaces=True,
                                       warnings=[])
    updater = Updater(ws_path=ws_path,
                      packages=workspace_packages,
                      conflict_strategy=conflict_strategy,
                      use_preprint=use_preprint,
                      num_threads=num_threads)
    updater.update_packages(packages)
    return 0


def fetch(packages,
          workspace,
          context,
          default_url,
          use_preprint,
          num_threads):
    """Fetch dependencies of a package.

    Args:
        packages (list): A list of packages provided by the user.
        workspace (str): Path to a workspace (without src/ in the end).
        context (Context): Current context. Needed to find current packages.
        default_url (str): A default url with a {package} placeholder in it.
        use_preprint (bool): Show status messages while cloning

    Returns:
        int: Return code. 0 if success. Git error code otherwise.
    """
    fetch_all = False
    if not packages:
        fetch_all = True

    ws_path = path.join(workspace, 'src')
    ignore_pkgs = Tools.list_all_ros_pkgs()

    already_fetched = set()
    packages = set(packages)

    global_error_code = Downloader.NO_ERROR

    # loop until there are no new dependencies left to download
    while(True):
        log.info(" Searching for dependencies.")
        deps_to_fetch = {}
        workspace_packages = find_packages(
            context.source_space_abs,
            exclude_subspaces=True, warnings=[])
        available_pkgs = [pkg.name for _, pkg in workspace_packages.items()]
        initial_cloned_pkgs = len(already_fetched)
        for package_path, package in workspace_packages.items():
            if package.name in already_fetched:
                continue
            if fetch_all or (package.name in packages):
                parser = Parser(default_url, package.name)
                package_folder = path.join(ws_path, package_path)
                deps_to_fetch = Tools.update_deps_dict(
                    deps_to_fetch, parser.get_dependencies(package_folder))
                if deps_to_fetch is None:
                    sys.exit(1)
                already_fetched.add(package.name)
                for new_dep_name in deps_to_fetch.keys():
                    # make sure we don't stop until we analyzed all
                    # dependencies as we have just added these repositories
                    # we must analyze their dependencies too even if we wanted
                    # to download dependencies for one project only.
                    packages.add(new_dep_name)
        try:
            downloader = Downloader(ws_path=ws_path,
                                    available_pkgs=available_pkgs,
                                    ignore_pkgs=ignore_pkgs,
                                    use_preprint=use_preprint,
                                    num_threads=num_threads)
        except ValueError as e:
            log.critical(" Encountered error. Abort.")
            log.critical(" Error message: %s", e.message)
            return 1
        error_code = downloader.download_dependencies(deps_to_fetch)
        if len(already_fetched) == initial_cloned_pkgs:
            log.info(" No new dependencies. Done.")
            break
        if error_code != 0:
            global_error_code = error_code
        log.info(" New packages available. Process their dependencies now.")
    return global_error_code



"""Hosts a class used to download dependencies.

Attributes:
    log (logging.Log): logger
"""
import logging

from os import path
from termcolor import colored
from concurrent import futures

from catkin_tools_fetch.lib.tools import Tools
from catkin_tools_fetch.lib.tools import GitBridge
from catkin_tools_fetch.lib.printer import Printer

log = logging.getLogger('deps')


class Downloader(object):
    """Downloader for dependency dict.

    Attributes:
        available_pkgs (str[]): dict of available packages in workspace
        ignore_pkgs (set): a set of packages to ignore (mostly ROS ones).
        ws_path (str): Workspace path. This is where packages live.
    """

    IGNORE_TAG = colored("[IGNORED]", 'yellow')
    NOT_FOUND_TAG = colored("[NOT FOUND]", 'red')
    CLONING_TAG = "[CLONING]"
    CHECKING_TAG = "[CHECKING]"

    NO_ERROR = 0

    def __init__(self,
                 ws_path,
                 available_pkgs,
                 ignore_pkgs,
                 use_preprint=True,
                 num_threads=4):
        """Init a downloader.

        Args:
            ws_path (str): Workspace path. This is where packages live.
            available_pkgs (iterable): dict of available packages in workspace.
            ignore_pkgs (iterable): set of packages to ignore (e.g. ROS ones).
        """
        super(Downloader, self).__init__()
        if not path.exists(ws_path):
            raise ValueError("""
            Folder '{}' is missing.
            Are you running 'fetch' from a catkin workspace?
            """.format(ws_path))
        self.ws_path = ws_path
        self.available_pkgs = available_pkgs
        self.ignore_pkgs = ignore_pkgs
        self.thread_pool = futures.ThreadPoolExecutor(max_workers=num_threads)
        self.use_preprint = use_preprint
        self.printer = Printer()

    def download_dependencies(self, dep_dict):
        """Check and download dependencies from a dependency dictionary.

        Args:
            dep_dict (dict): dictionary {name: dep} with dependencies.

        Returns:
            int: Return code. 0 if all fine. Git error code otherwise.
        """
        if not isinstance(dep_dict, dict):
            raise ValueError("expected a dictionary with dependencies.")
        for dep in dep_dict.values():
            log.debug(" dep before: %s", dep)
        checked_deps = self.__check_dependencies(dep_dict)
        return self.__clone_dependencies(checked_deps)

    def __clone_dependency(self, pkg_name, url, dep_path, branch):
        """Clone a single dependency. Return a future to the clone process."""
        if self.use_preprint:
            msg = " {}: {}".format(Tools.decorate(pkg_name),
                                   Downloader.CLONING_TAG)
            self.printer.add_msg(pkg_name, msg)
        return GitBridge.clone(pkg_name, url, dep_path, branch)

    def __clone_dependencies(self, checked_deps):
        """Clone dependencies.

        Args:
            checked_deps (dict): Dict {name: dep} with valid dependencies.

        Returns:
            int: Error code. 0 if all fine, Result from git error otherwise.
        """
        if not checked_deps:
            # exit early if there is nothing to download
            return Downloader.NO_ERROR
        log.info(" Cloning valid dependencies:")
        error_code = Downloader.NO_ERROR
        # store all tasks in a futures list
        futures_list = []
        for name, dependency in checked_deps.items():
            url = dependency.url
            branch = dependency.branch
            log.debug(" prepare clone: url: %s, branch: %s", url, branch)
            if not branch:
                branch = "master"
            if name in self.available_pkgs:
                msg = " {}: {}".format(
                    Tools.decorate(name), GitBridge.EXISTS_TAG)
                self.printer.purge_msg(name, msg)
                continue
            dep_path = path.join(self.ws_path, name)
            future = self.thread_pool.submit(
                self.__clone_dependency, name, url, dep_path, branch)
            futures_list.append(future)
        # we have all the futures ready. Now just wait for them to finish.
        for future in futures.as_completed(futures_list):
            pkg_name, clone_result = future.result()
            msg = " {}: {}".format(
                Tools.decorate(pkg_name), clone_result)
            self.printer.purge_msg(pkg_name, msg)
            if clone_result == GitBridge.ERROR_TAG:
                error_code = 1
        return error_code

    def __check_dependency(self, dependency):
        if self.use_preprint:
            msg = " {}: {}".format(
                Tools.decorate(dependency.name), Downloader.CHECKING_TAG)
            self.printer.add_msg(dependency.name, msg)
        return GitBridge.repository_exists(dependency)

    def __check_dependencies(self, dep_dict):
        """Check dependencies for validity.

        We don't want to avoid packages that we ignore or those which
        repositories do not exist.

        Args:
            dep_dict (dict): A dictionary {name: dep} with dependencies.

        Returns:
            dict: Only valid dependencies from the input dict.
        """
        checked_deps = {}
        if not dep_dict:
            # exit early if there are no new dependencies
            return checked_deps
        futures_list = []
        log.info(" Checking merged dependencies:")
        for dependency in dep_dict.values():
            log.debug(" check dep: %s", dependency)
            if dependency.name in self.ignore_pkgs:
                msg = " {}: {}".format(
                    Tools.decorate(dependency.name), Downloader.IGNORE_TAG)
                self.printer.add_msg(dependency.name, msg)
                continue
            futures_list.append(self.thread_pool.submit(
                self.__check_dependency, dependency))
        for future in futures.as_completed(futures_list):
            dependency, repo_found = future.result()
            if repo_found:
                msg = " {}: {}".format(
                    Tools.decorate(dependency.name), dependency.url)
                self.printer.purge_msg(dependency.name, msg)
                checked_deps[dependency.name] = dependency
            else:
                msg = " {}: {}".format(
                    Tools.decorate(dependency.name), Downloader.NOT_FOUND_TAG)
                self.printer.purge_msg(dependency.name, msg)
        return checked_deps



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .cli import main
from .cli import prepare_arguments
from .cli import prepare_arguments_deps

# This describes this command to the loader
description = dict(
    verb='fetch',
    description='Fetch dependencies of packages in the workspace',
    main=main,
    prepare_arguments=prepare_arguments,
)

# This describes this command to the loader
description_deps = dict(
    verb='deps',
    description='Manage dependencies',
    main=main,
    prepare_arguments=prepare_arguments_deps,
)

# This describes this command to the loader
description_dependencies = dict(
    verb='dependencies',
    description='Manage dependencies',
    main=main,
    prepare_arguments=prepare_arguments_deps,
)
